#!/usr/bin/env python

import weakref as _weakref
import logging as _logging

import notify2 as _notify

import gi as _gi
_gi.require_version('Gtk', '3.0')
from gi.repository import Gtk as _gtk
_gi.require_version('Gdk', '3.0')
from gi.repository import Gdk as _gdk
from gi.repository import GObject as _gobject
_gi.require_version('Keybinder', '3.0')
from gi.repository import Keybinder as _keybinder

import dbus as _dbus

import klemmbrett.util as _util
import klemmbrett.config as _config


_log = _logging.getLogger(__name__)


class KlemmbrettVormKopf(Exception):
    pass


class Klemmbrett(_gobject.GObject):

    __gsignals__ = {
        "text-selected": (_gobject.SIGNAL_RUN_FIRST, None, (_gobject.TYPE_PYOBJECT,)),
        "text-set": (_gobject.SIGNAL_RUN_FIRST, None, (_gobject.TYPE_PYOBJECT,)),
    }

    _PLUGIN_PREFIX = "plugin "
    _TIE_PREFIX = "tie:"

    def __init__(self, config_files):
        super(Klemmbrett, self).__init__()
        self._clipboard = _gtk.Clipboard.get(_gdk.SELECTION_CLIPBOARD)
        self._primary = _gtk.Clipboard.get(_gdk.SELECTION_PRIMARY)

        self.config = _config.Config()
        self.config.read(config_files)

        _keybinder.init()

        # configure klemmbrett
        self._plugins = dict()
        self._load_plugins()

        self._sync = _util.humanbool(self.config.get('klemmbrett', 'sync', True))

        self.selection = None

        self._clipboard.connect('owner-change', self._clipboard_owner_changed)
        self._primary.connect('owner-change', self._clipboard_owner_changed)

        try:
            _notify.init("Klemmbrett")
        except _dbus.exceptions.DBusException:
            _log.error("Could not register with notification interface, notifications will not work properly", exc_info=True)

    def _load_plugins(self):
        for section in self.config.sections():
            if not section.startswith(self._PLUGIN_PREFIX):
                continue

            name = section[len(self._PLUGIN_PREFIX):].strip()
            opts = dict(self.config.items(section))
            # load the plugin module and create an instance
            plugin = _util.load_dotted(opts['plugin'])
            plugin = plugin(
                name,
                dict(plugin.OPTIONS, **opts),
                self,
            )
            self._plugins[name] = plugin


        # inject attributes into plugins on a per plugin config
        # basis via config options of the form:
        # tie:plugin_local_identitier = global_name
        # or from the plugin internal config dict.
        for name, plugin in self._plugins.items():
            for key, value in plugin.options.items():
                if not key.startswith(self._TIE_PREFIX):
                    continue

                source = key[len(self._TIE_PREFIX):].strip()
                target = _weakref.proxy(self._plugins[value])
                setattr(self._plugins[name], source, target)

        # only after alll plugins are instantiated and properly
        # injected, they may connect to other plugins signals,
        # so we, sadly, have to live with some kind of bootstrap
        # method with some additional code
        for plugin in self._plugins.values():
            plugin.bootstrap()

    def _clipboard_owner_changed(self, clipboard, event):
        text = clipboard.wait_for_text()

        if text != self.selection and text is not None:
            self.selection = text
            if self._sync:
                if clipboard is self._primary:
                    self._clipboard.set_text(text, -1)
                elif clipboard is self._clipboard:
                    self._primary.set_text(text, -1)

            self.emit("text-selected", text)

        return True

    def notify(self, summary, text):
        """ Display a message about the new suggestion and its origin """
        n = _notify.Notification(summary, text)
        n.show()

    def set(self, text, primary = True, clipboard = True):
        try:
            if clipboard:
                self._clipboard.set_text(text, -1)
            if primary:
                self._primary.set_text(text, -1)
            self.emit("text-set", text)
        except TypeError:
            # TypeError: Gtk.Clipboard.set_text() argument 1 must be string, not None
            # FIXME(mbra): i really do not know when and when this happens :/
            pass

    def main(self):
        _gtk.main()





#!/usr/bin/env python

import html as _html
import pkg_resources as _pr
import distutils.util as _util

try:
    from compiler.consts import CO_GENERATOR
except ImportError:
    # IronPython doesn't have a complier module
    CO_GENERATOR = 0x20


def get_status_icon_filename():
    return _pr.resource_filename("klemmbrett", "assets/klemmbrett.png")


def isgenerator(func):
    """ Check if a given function is a generator before calling it """
    try:
        return func.func_code.co_flags & CO_GENERATOR != 0
    except AttributeError:
        return False


def yieldwrap(func, *args, **kwargs):
    """ Ensure that the toplevel function is seen as a generator """
    def wrapped():
        for i in func(*args, **kwargs):
            yield i
    return wrapped


def humanbool(value):
    """
        Use distutils.util strtobool to convert various boolean tokens
        like yes, on, no, off etc to a boolean
    """
    return _util.strtobool(str(value).strip().lower() or 'no')


def load_dotted(name):
    """ Imports and return the given dot-notated python object """
    components = name.split('.')
    path = [components.pop(0)]
    obj = __import__(path[0])
    while components:
        comp = components.pop(0)
        path.append(comp)
        try:
            obj = getattr(obj, comp)
        except AttributeError:
            __import__('.'.join(path))
            try:
                obj = getattr(obj, comp)
            except AttributeError:
                raise ImportError('.'.join(path))
    return obj


def htmlsafe(text):
    """ Escape htmlentities """
    return _html.escape(text)




# coding: utf-8
import os as _os

import gi as _gi
_gi.require_version('Gtk', '3.0')
from gi.repository import Gtk as _gtk
_gi.require_version('AppIndicator3', '0.1')
from gi.repository import AppIndicator3 as _appindicator
from klemmbrett import plugins as _plugins

import klemmbrett.util as _util
import klemmbrett.about as _about


class AppIndicatorPlugin(_plugins.Plugin):

    def __init__(self, *args, **kwargs):
        super(AppIndicatorPlugin, self).__init__(*args, **kwargs)

        icon = self.options.get("icon-path", _util.get_status_icon_filename())
        self.indicator = _appindicator.Indicator.new(
            "klemmbrett",
            icon,
            _appindicator.IndicatorCategory.APPLICATION_STATUS,
        )
        self.indicator.set_icon_full(icon, "Klemmbrett")
        self.indicator.set_status(_appindicator.IndicatorStatus.ACTIVE)

        self.menu = _gtk.Menu()

        item = _gtk.MenuItem("Quit")
        item.connect('activate', _gtk.main_quit)
        self.menu.append(item)

        item = _gtk.MenuItem("About")
        item.connect('activate', _about.about)
        self.menu.append(item)

        self.menu.show_all()

        self.indicator.set_menu(self.menu)



#!/usr/bin/env python

import configparser as _config

undefined = object()

class Config(_config.RawConfigParser):

    def get(self, section, option, default = undefined):
        try:
            return _config.RawConfigParser.get(self, section, option)
        except (_config.NoSectionError, _config.NoOptionError):
            if default is not undefined:
                return default
            raise





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os as _os
import sha as _sha
import hmac as _hmac
import time as _time
import functools as _ft
import socket as _socket
import xmlrpclib as _xmlrpc
import urlparse as _urlparse
import threading as _threading
import SimpleXMLRPCServer as _xmlrpcserver

import gi as _gi
_gi.require_version('Gdk', '3.0')
from gi.repository import Gdk as _gdk
from gi.repository import GObject as _gobject
_gi.require_version('Keybinder', '3.0')
from gi.repository import Keybinder as _keybinder

import Crypto.Cipher.AES as _aes

import klemmbrett.util as _util
import klemmbrett.plugins as _plugins


def hosttuple(host):
    parts = host.split(":")
    if len(parts) == 2:
        host, port = parts
    if len(parts) == 1:
        host, port = parts[0], 6789

    port = int(port)

    return (host, port)


class ClipboardExchangeHandler(_xmlrpcserver.SimpleXMLRPCRequestHandler):
    """
        The ClipboardExchangeHandler is used to inject the client_address
        into the method call of all exported xmlrpc methods, as well as
        decoding/verifing the xmlrpc encryption.
    """

    def __init__(self, destinations, *args, **kwargs):
        self._destinations = destinations
        _xmlrpcserver.SimpleXMLRPCRequestHandler.__init__(self, *args, **kwargs)

    def _dispatch(self, method, params):
        """ Inject the client_address into the argument list of all rpc methods """
        f = list(params)
        f.append(self.client_address)
        return self.server._dispatch(method, f)

    def decode_request_content(self, data):
        """ Do decryption/verification of the rpc request """
        try:
            hmac, issued, data = data.split("|")
            issued = int(issued)

            data = data.decode("hex")
            hmac = hmac.decode("hex")

            h = str(
                _hmac.new(
                    str(self._destinations[self.client_address[0]]["hmac-key"]),
                    msg = data,
                    digestmod = _sha,
                ).digest()
            )

            if h != hmac:
                self.send_response(400, "Request Verification failed")
                self.send_header("Content-length", "0")
                self.end_headers()
                return None


            # the first block_size bytes of the message is the initialization vector
            data = _aes.new(
                self._destinations[self.client_address[0]]["encryption-key"].decode("hex"),
                _aes.MODE_CFB,
                data[:_aes.block_size],
            ).decrypt(data[_aes.block_size:])

            # only accept messages in a narrow timeslot, best whould be
            # to have a logbook of request ids used or some sort of sequence
            # but that seems excessive at the moment, on the other hand this
            # requires the systems to sync their time, either by using a
            # timeserver or some sort of internal protocol
            now = _time.time()
            if issued < now - 5:
                self.send_response(400, "The request is no longer valid")
                self.send_header("Content-length", "0")
                self.end_headers()
                return None

            return data
        except:
            import traceback as _tb
            _tb.print_exc()

    def do_POST(self):
        """Handles the HTTP POST request.

        Attempts to interpret all HTTP POST requests as XML-RPC calls,
        which are forwarded to the server's _dispatch method for handling.
        """

        # Check that the path is legal
        if not self.is_rpc_path_valid():
            self.report_404()
            return

        try:
            # Get arguments by reading body of request.
            # We read this in chunks to avoid straining
            # socket.read(); around the 10 or 15Mb mark, some platforms
            # begin to have problems (bug #792570).
            max_chunk_size = 10*1024*1024
            size_remaining = int(self.headers["content-length"])
            L = []
            while size_remaining:
                chunk_size = min(size_remaining, max_chunk_size)
                L.append(self.rfile.read(chunk_size))
                size_remaining -= len(L[-1])
            data = self.decode_request_content(''.join(L))

            # In previous versions of SimpleXMLRPCServer, _dispatch
            # could be overridden in this class, instead of in
            # SimpleXMLRPCDispatcher. To maintain backwards compatibility,
            # check to see if a subclass implements _dispatch and dispatch
            # using that method if present.
            response = self.server._marshaled_dispatch(
                data,
                getattr(self, '_dispatch', None),
            )
        except Exception as e: # This should only happen if the module is buggy
            # internal error, report as HTTP server error
            self.send_response(500)

            # Send information about the exception if requested
            if hasattr(self.server, '_send_traceback_header') and \
                    self.server._send_traceback_header:
                self.send_header("X-exception", str(e))
                self.send_header("X-traceback", traceback.format_exc())

            self.end_headers()
        else:
            # got a valid XML RPC response
            self.send_response(200)
            self.send_header("Content-type", "text/xml")
            self.send_header("Content-length", str(len(response)))
            self.end_headers()
            self.wfile.write(response)

            # shut down the connection
            self.wfile.flush()
            self.connection.shutdown(1)


class ClipboardExchangeTransport(_xmlrpc.Transport):
    """
        The ClipboardExchangeTransport is only used for encoding/signing
        the xmlrpc request
    """

    def __init__(self, encryption_key, hmac_key, use_datetime = 0):
        """ Store encryption and hmac key for later """
        _xmlrpc.Transport.__init__(self, use_datetime)
        self._encryption_key = encryption_key
        self._hmac_key = hmac_key

    def request(self, host, handler, request_body, verbose=0):
        """
            Intercept the request, hash it, encrypt it, add a timestampt
            and transmit it with the standard transports request method
        """
        encryption_iv = _os.urandom(_aes.block_size)
        issued = int(_time.time())

        request_body = encryption_iv + _aes.new(
            self._encryption_key.decode('hex'),
            _aes.MODE_CFB,
            encryption_iv,
        ).encrypt(request_body)
        h = _hmac.new(self._hmac_key, msg = request_body, digestmod = _sha).digest()

        return _xmlrpc.Transport.request(
            self,
            host,
            handler,
            '|'.join([
                h.encode('hex'),
                str(issued),
                request_body.encode('hex'),
            ]),
            verbose,
        )


class ClipboardExchange(_plugins.PopupPlugin):

    DEFAULT_BINDING = "<Ctrl><Alt>P"
    DEFAULT_ACCEPT_BINDING = "<Ctrl><Alt>X"
    DEFAULT_USERHISTORY_BINDING = "<Ctrl><Alt>D"
    OPTIONS = {
        "tie:history": "history",
    }

    def get_destinations(self):
        """
            Create a hash of all configured destinations, indexed
            by ip as a string, so we can look it up easily in the
            RequestHandler for injection into the argument list.

            Each hash contains the rpc url, the name of the user,
            a history object for that user as well as the encryption
            and hmac keys for this user, at the moment this keys
            are the same for all users, but this will change in
            the future.
        """
        dests = dict()

        for name, addr in self.options.items():
            if not name.startswith('user.'):
                continue

            host, port = hosttuple(addr)
            ip = _socket.gethostbyname(host)
            url = "http://%s:%s" % (host, port)

            dests[ip] = {
                "name": name[5:],
                "url": url,
                "history": _plugins.HistoryController("history." + name[5:], self.options, self.klemmbrett),
                "encryption-key": self.options["encryption-key"],
                "hmac-key": self.options["hmac-key"],
            }

            for ip, dest in dests.items():
                dest["history"].bootstrap()

        return dests

    def bootstrap(self):
        """
            Setup all destinations and shortcuts and start the rpc server
        """
        self._destinations = self.get_destinations()

        # initialize standard stuff
        #_notify.init("Klemmbrett")
        self._current_suggestion = None

        # binding to accept the suggested text into the clipboard
        _keybinder.bind(
            self.options.get('accept-suggestion-shortcut', self.DEFAULT_ACCEPT_BINDING),
            self._accept_suggestion,
        )

        _keybinder.bind(
            self.options.get('user-history-shortcut', self.DEFAULT_USERHISTORY_BINDING),
            self._show_histories,
        )

        _plugins.PopupPlugin.bootstrap(self)
        self._start_server()

    def _show_histories(self):
        """ Popup a dropdown menu with a submenu per user containing their history """
        return self.popup(((x["name"], x["history"].items) for x in self._destinations.values()))

    def _serve(self):
        """ Middleman for the Thread ctors target argument """
        return self._server.serve_forever()

    def _start_server(self):
        """
            Initialize the xmlrpc server, register all required functions
            and run the server in a seperate thread
        """
        host, port = hosttuple(self.options.get("listen", "0.0.0.0"))

        _gdk.threads_init()
        self._server = _xmlrpcserver.SimpleXMLRPCServer(
            (host, port),
            allow_none = True,
            requestHandler = _ft.partial(
                ClipboardExchangeHandler,
                self._destinations,
            ),
        )
        self._server.register_function(self._suggest, "suggest")
        #self._server.register_introspection_functions()

        t = _threading.Thread(target = self._serve)
        t.daemon = True
        t.start()

    def _accept_suggestion(self):
        """ Accept the current suggestion into the main clipboard history """
        self.history.add(self._current_history.top)
        self.klemmbrett.set(self._current_history.top)

    def _suggest(self, text, client_address):
        """ Display a message about the new suggestion and it origin """
        _gdk.threads_enter()
        self._destinations[client_address[0]]["history"].add(text)
        self._current_history = self._destinations[client_address[0]]["history"]
        self.klemmbrett.notify(
            "Suggestions from %r" % (
                self._destinations[client_address[0]]["name"],
            ),
            self._printable(text, htmlsafe = True),
        )
        _gdk.threads_leave()

    def items(self):
        """
            The default item provider for PopupPlugins. This will display
            the list of destinations to send the current clipboard contents to.
        """
        for dest in self._destinations.values():
            yield (
                dest["name"],
                _ft.partial(self._send_text, dest["url"], self.history.top),
            )

    def _send_text(self, dest, text):
        """
            Callback for the items method. sends a suggestion to the desired
            destination
        """
        p =  _xmlrpc.ServerProxy(
            dest,
            transport = ClipboardExchangeTransport(
                self.options["encryption-key"],
                self.options["hmac-key"],
            ),
        )
        p.suggest(text)



# coding: utf-8

import gi as _gi
_gi.require_version('Gtk', '3.0')
from gi.repository import Gtk as _gtk
from gi.repository import GdkPixbuf as _pixbuf


import klemmbrett.util as _util

def about(event):
    dialog = _gtk.AboutDialog()
    # lists of authors and documenters (will be used later)
    authors = [
        "Michael van Bracht",
        "Roland Sommer",
        "Sven Wegener",
        "Marc Schmitzer",
    ]

    # we fill in the aboutdialog
    dialog.set_program_name("Klemmbrett")
    dialog.set_authors(authors)
    dialog.set_artists(["Tray icon provided by icons8.com https://icons8.com"])
    dialog.set_website("https://github.com/wontfix-org/klemmbrett")
    dialog.set_copyright("Tray Icon provided by https://icons8.com")
    dialog.set_website_label("Klemmbrett on GitHub")
    dialog.set_license_type(_gtk.License.MIT_X11)
    logo = _pixbuf.Pixbuf.new_from_file(_util.get_status_icon_filename())
    dialog.set_logo(logo)

    # the program name
    dialog.set_title("")

    # to close the aboutdialog when "close" is clicked we connect the
    # "response" signal to the destroy action
    dialog.connect("response", lambda action, parameter: action.destroy())

    # show the aboutdialog
    dialog.show()



#!/usr/bin/env python

import os as _os
import re as _re
import itertools as _it
import functools as _ft
import pickle as _pickle
import weakref as _weakref
import logging as _logging
import collections as _collections

import gi as _gi

_gi.require_version('Gtk', '3.0')
from gi.repository import Gtk as _gtk

_gi.require_version('Gdk', '3.0')
from gi.repository import Gdk as _gdk
from gi.repository import GdkX11 as _gdkx11
from gi.repository import GObject as _gobject

_gi.require_version('Keybinder', '3.0')
from gi.repository import Keybinder as _keybinder

import klemmbrett.util as _util
import klemmbrett as _klemmbrett
import klemmbrett.about as _about
import klemmbrett.config as _config

_log = _logging.getLogger(__name__)


class HistoryEmpty(Exception):
    pass


class Plugin(_gobject.GObject):

    OPTIONS = {}

    def __init__(self, name, options, klemmbrett):
        if hasattr(self, '_initialized'):
            return

        _gobject.GObject.__init__(self)

        if type(klemmbrett) in _weakref.ProxyTypes:
            self.klemmbrett = _weakref.proxy(klemmbrett)
        else:
            self.klemmbrett = klemmbrett

        self.options = options
        self.name = name

    def set(self, widget = None, text = None):
        if callable(text):
            text = text()
        self.klemmbrett.set(text)

    def _printable(self, text, htmlsafe = False):
        ll = int(self.options.get('line-length', 30))
        clean = _re.sub(r'\s+', ' ', text).strip()

        if len(clean) > ll:
            om = self.options.get('omit-mode', 'middle')
            if om == 'start':
                clean = clean[len(clean) - ll:]
            elif om == 'middle':
                clean = clean[:int(ll / 2)] + " ... " + clean[len(clean) - int(ll / 2):]
            elif om == 'end':
                clean = clean[:ll]

        if htmlsafe:
            clean = _util.htmlsafe(clean)

        return clean

    def bootstrap(self):
        pass


class StatusIcon(Plugin):

    def __init__(self, *args, **kwargs):
        super(StatusIcon, self).__init__(*args, **kwargs)

        self.menu = _gtk.Menu()

        item = _gtk.MenuItem("About")
        item.connect('activate', _about.about)
        self.menu.append(item)

        item = _gtk.MenuItem("Quit")
        item.connect('activate', _gtk.main_quit)
        self.menu.append(item)

        self.menu.show_all()

        self.tray = _gtk.StatusIcon()
        self.tray.set_visible(True)
        self.tray.set_tooltip_text("Klemmbrett")
        self.tray.connect('popup-menu', self.on_menu, self.menu)

        icon = self.options.get('icon-path', _util.get_status_icon_filename())
        if _os.path.sep in icon:
            self.tray.set_from_file(_os.path.expanduser(icon))
        else:
            self.tray.set_from_icon_name(icon)

    def on_menu(self, icon, event_button, event_time, menu):
        menu.popup(
            None,
            None,
            _gtk.StatusIcon.position_menu,
            self.tray,
            event_button,
            event_time,
        )


class PopupPlugin(Plugin):

    def bootstrap(self):
        _keybinder.bind(
            self.options.get('shortcut', self.DEFAULT_BINDING),
            self.popup,
        )

    def _expand(self, widget, iterable):
        sm = widget.get_submenu()
        self._build_menu(sm, iterable())
        sm.show_all()

    def _build_menu(self, menu, items):
        accels = list(
            _it.chain(
                range(0, 10),
                map(
                    chr,
                    _it.chain(
                        range(ord('a'), ord('z') + 1),
                        range(ord('A'), ord('Z') + 1)),
                )
            )
        )
        for pos, (label, value) in enumerate(items):
            try:
                label = "_%s %s" % (accels[pos], label.replace('_', '__'))
                item = _gtk.MenuItem(label, use_underline = True)
            except IndexError:
                # We ran out of accelerator keys, too bad...
                item = _gtk.MenuItem(label, use_underline = False)

            if _util.isgenerator(value):
                item.set_submenu(_gtk.Menu())
                item.connect("activate", self._expand, value)
            else:
                item.connect("activate", self.set, value)

            menu.append(item)

    def _ungrab_keyboard(self):
        dm = _gdkx11.X11DeviceManagerCore(display=_gdk.Display.get_default())
        for dev in dm.list_devices(_gdk.DeviceType.MASTER):
            if dev.get_source() == _gdk.InputSource.KEYBOARD:
                dev.ungrab(_keybinder.get_current_event_time())

    def popup(self, keystr, items = None):
        self._ungrab_keyboard()

        menu = _gtk.Menu()
        index = 0

        if _util.humanbool(self.options.get('show-current-selection', 'yes')) and len(self.history):
            item = _gtk.MenuItem("")
            item.get_children()[0].set_markup(
                "<b>%s</b>" % (self._printable(self.history.top, True),),
            )
            menu.append(item)
            menu.append(_gtk.SeparatorMenuItem())
            index += 1

        self._build_menu(menu, items or self.items())

        menu.show_all()
        menu.popup(
            None,
            None,
            None,
            None,
            0,
            _keybinder.get_current_event_time(),
        )
        menu.set_active(index)
        return True


class FancyItemsMixin(object):

    SIMPLE_SECTION_DEFAULT = "value"
    _TYPE_SEPERATOR = "."

    def bootstrap(self):
        self._items = []
        section = self.options.get('simple-section', self.SIMPLE_SECTION)
        default_type = self.options.get('simple-section-default', self.SIMPLE_SECTION_DEFAULT)

        if not self.klemmbrett.config.has_section(section):
            raise KeyError("No config section %s defined" % (section,))

        for label, value in self.klemmbrett.config.items(section):
            target = default_type
            if self._TYPE_SEPERATOR in label:
                target, label = label.split(self._TYPE_SEPERATOR, 1)

            self._items.append((label, {target: value}))

        prefix = self.options.get('complex-section-prefix', self.COMPLEX_SECTION_PREFIX)

        for section in self.klemmbrett.config.sections():
            if not section.startswith(prefix):
                continue

            options = dict(self.klemmbrett.config.items(section))
            label = section[len(prefix):]

            self._items.append((label, options))

            if "shortcut" in options:
                _keybinder.bind(
                    options['shortcut'],
                    _ft.partial(self.set, widget = None, text = label),
                )

    def items(self):
        return iter(self._items)


class HistoryController(Plugin):

    __gsignals__ = {
        "text-accepted": (_gobject.SIGNAL_RUN_FIRST, None, (_gobject.TYPE_PYOBJECT,)),
    }

    def __init__(self, name, options, klemmbrett):
        Plugin.__init__(self, name, options, klemmbrett)
        self._history = _collections.deque(maxlen = self.maxlen)
        self._extend_detection = _util.humanbool(self.options.get('extend-detection', 'yes'))

    def items(self):
        for text in self._history:
            yield (
                self._printable(text),
                text,
            )

    def is_extended(self, text):
        if not self._extend_detection:
            return False

        if (
            len(self)
            and (
                text.startswith(self.top)
                or text.endswith(self.top)
            )
        ):

            return True
        return False

    def add(self, text, emit = True):
        if self.accepts(text):
            if self.is_extended(text):
                self._history[0] = text
            else:
                self._history.appendleft(text)

            if emit:
                self.emit("text-accepted", text)
            return True
        return False

    def __iter__(self):
        return iter(self._history)

    def __len__(self):
        return len(self._history)

    def accepts(self, text):
        # do not accept bullshit
        if not isinstance(text, str):
            return False

        # do not accept empty strings and pure whitespace strings
        text = text.strip()
        if not text:
            return False

        # accept everything if the history is empty
        if not len(self._history):
            return True

        # only if it is not the current selection
        return text != self.top

    @property
    def top(self):
        if not self._history:
            raise HistoryEmpty("The history is empty")
        return self._history[0]

    @property
    def maxlen(self):
        return int(self.options.get("length", 15))


class HistoryPicker(HistoryController, PopupPlugin):

    DEFAULT_BINDING = "<Ctrl><Alt>C"

    OPTIONS = {
        "tie:history": "history",
    }

    def __init__(self, name, options, klemmbrett):
        PopupPlugin.__init__(self, name, options, klemmbrett)
        HistoryController.__init__(self, name, options, klemmbrett)

    def bootstrap(self):
        PopupPlugin.bootstrap(self)
        HistoryController.bootstrap(self)
        self.klemmbrett.connect("text-selected", self._text_selected)

    def _text_selected(self, widget, text):
        return self.add(text)


class PersistentHistory(Plugin):
    OPTIONS = {
        "tie:history": "history",
    }

    def __init__(self, *args, **kwargs):
        super(PersistentHistory, self).__init__(*args, **kwargs)
        self._histfile = _os.path.expanduser(self.options.get("histfile", "~/.klemmbrett.history"))

    def bootstrap(self):
        self._load()
        self._persist = open(self._histfile, "ba")
        self.history.connect("text-accepted", self._text_accepted)

    def _load(self):
        if not _os.path.exists(self._histfile):
            return

        history = open(self._histfile, "br")
        dq = _collections.deque(maxlen = self.history.maxlen)

        while True:
            try:
                dq.append(_pickle.load(history))
            except EOFError:
                break

        history.close()

        for i in dq:
            self.history.add(i, False)

    def _text_accepted(self, widget, text):
        # FIXME(mbra): we need to handle a history size limit, so the file does
        # not grow indefinitely
        _pickle.dump(text, self._persist, protocol = _pickle.HIGHEST_PROTOCOL)
        self._persist.flush()
        return True


class MultiPicker(PopupPlugin, FancyItemsMixin):

    SIMPLE_SECTION_DEFAULT = "value"
    OPTIONS = {
        "tie:history": "history"
    }

    def bootstrap(self):
        PopupPlugin.bootstrap(self)
        FancyItemsMixin.bootstrap(self)
        self._types = (
            ("value", self._value),
            ("callable", self._callable),
            ("action", self._action),
            ("notify", self._notify),
        )

    def items(self):
        for label, options in FancyItemsMixin.items(self):
            callable = lambda: self.history.top

            for type, handler in self._types:
                if type in options:
                    callable = handler(label, options, callable)

            yield (label, callable)

    def _value(self, label, options, callable):
        return _ft.partial(options.get, "value")

    def _callable(self, label, options, callable):
        try:
            return _util.load_dotted(options["callable"])(options, self)
        except HistoryEmpty:
            _log.info("History is empty")
            self.klemmbrett.notify("The history is empty", "The history is empty")


    def _action(self, label, options, callable):
        try:
            return _ft.partial(self._perform_action, options, callable)
        except HistoryEmpty:
            _log.info("History is empty")
            self.klemmbrett.notify("The history is empty", "The history is empty")

    def _notify(self, label, options, callable):
        return _ft.partial(self.klemmbrett.notify, "Klemmbrett", options.get("notify"))

    @staticmethod
    def _perform_action(options, callable):
        _gobject.spawn_async([
            "/bin/bash",
            "-c",
            options["action"] % (callable(),),
        ]),

    def set(self, widget = None, text = None):
        return PopupPlugin.set(self, widget = widget, text = text)


class ActionPicker(MultiPicker):
    DEFAULT_BINDING = "<Ctrl><Alt>A"
    SIMPLE_SECTION = "actions"
    COMPLEX_SECTION_PREFIX= "action "
    SIMPLE_SECTION_DEFAULT = "action"


class SnippetPicker(MultiPicker):
    DEFAULT_BINDING = "<Ctrl><Alt>S"
    SIMPLE_SECTION = "snippets"
    COMPLEX_SECTION_PREFIX= "snippet "



# coding: utf-8
"""
===========================
DBus service for Klemmbrett
===========================

The `DBusServicePlugin` allows other applications to inject
things into your history via dbus.

You can activate this plugin by adding the following to your
`klemmbrett.conf` file:

::
    [plugin dbus]
    plugin = klemmbrett.plugins.dbus_service.DBusServicePlugin

If you want to inject things into your history you can use the
`Klemmbrett` class from this file like this:

    >>> from klemmbrett.plugins.dbus_service import Klemmbrett
    >>> klemmbrett = Klemmbrett()
    >>> klemmbrett.add("TEXT FOR CLIPBOARD")
"""

import dbus
import dbus.glib # backwards compatible setting of mainloop
import dbus.service

from klemmbrett import plugins as _plugins

DBUS_NAME = "org.wontfix.Klemmbrett"
DBUS_PATH = "/org/wontfix/Klemmbrett"


class Klemmbrett(object):
    """
    Simple helper class to get the DBus proxy object. All calls
    to an instance are proxied to the underlying dbus object.
    """
    def __init__(self):
        bus = dbus.SessionBus()
        self.proxy = bus.get_object(DBUS_NAME, DBUS_PATH)

    def __getattr__(self, name):
        return getattr(self.proxy, name)


class DBusService(dbus.service.Object):
    """
    DBus service class for klemmbrett.

    This had to go to a seperate class to solve the famous
    `TypeError: metaclass conflict` you get if you inherit from
    `klemmbrett.plugins.Plugin` and `dbus.service.Object`.
    """
    def __init__(self, plugin):
        self.plugin = plugin
        bus_name = dbus.service.BusName(DBUS_NAME, bus=dbus.SessionBus())
        dbus.service.Object.__init__(self, bus_name, DBUS_PATH)

    @dbus.service.method(dbus_interface=DBUS_NAME, in_signature='s')
    def add(self, text):
        self.plugin.add(text)


class DBusServicePlugin(_plugins.Plugin):
    OPTIONS = {
        "tie:history": "history",
    }

    def __init__(self, name, options, klemmbrett):
        _plugins.Plugin.__init__(self, name, options, klemmbrett)
        self.dbus = DBusService(self)

    def add(self, text):
        self.history.set(text = text)




import re as _re
import os as _os

import klemmbrett.util as _util


def newline_to_comma_quoted(options, plugin):
    def n2cq():
        try:
            text = plugin.history.top
        except IndexError:
            return None
        return "'" + "','".join(_re.split(r"\s+", text.strip())) + "'"
    return n2cq


def newline_to_comma(options, plugin):
    def n2c():
        try:
            text = plugin.history.top
        except IndexError:
            return None
        return text.strip().replace("\n", ",")
    return n2c


def fswalker(options, plugin):
    def walk(base = None):
        if base is None:
            base = options.get("base", _os.getenv("HOME"))

        for item in _os.listdir(base):
            if item.startswith("."):
                continue

            path = _os.path.join(base, item)
            if _os.path.isfile(path):
                yield (item, path)
            elif _os.path.isdir(path):
                yield (item, _util.yieldwrap(walk, path))

    return walk



#!/usr/bin/env python

import setuptools as _setuptools

_setuptools.setup(
    name = "klemmbrett",
    version = "0.3.3",
    description = "Versatile clipboard manager",
    author = "Michael van Bracht",
    author_email = "michael@wontfix.org",
    url = "https://github.com/wontfix-org/klemmbrett",
    license="MIT",
    packages = _setuptools.find_packages(),
    scripts = ["scripts/klemmbrett"],
    # XXX(mvb): This is probably too os/distribution dependent and
    # should be moved to the actual packaging code
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    data_files = [
        ("/etc", ["conf/klemmbrett.conf"]),
        ("/etc/xdg/autostart/", ["conf/klemmbrett-autostart.desktop"]),
        ("share/klemmbrett/", ["data/gtk-paste.svg", "data/gtk-paste2.svg"]),
#        ("share/applications/", ["conf/klemmbrett.desktop"]),
    ],
    include_package_data=True,
    install_requires = [
        "notify2",
        "dbus-python",
        "pygobject",
        "pycrypto",
    ],
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Environment :: X11 Applications :: GTK"
    ],
)




import sqlalchemy as _sa

def statement(options, plugin):
    def stmt():
        return str(
            list(
                _sa.create_engine(options["engine"]).execute(
                    _sa.sql.text(options["statement"]),
                    dict(
                        ((str(x), y) for x, y in enumerate(list(plugin.history)))
                    )
                )
            )[0][0]
        )
    return stmt
